/*
Copyright (C) 2009-2013 Baker

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// lists.c

#define CORE_LOCAL
#include "lists.h"
#include "core.h"

void List_Add (clist_t** headnode, const char *name)
{
	clist_t	*listent,*cursor,*prev;

	// ignore duplicate
	for (listent = *headnode; listent; listent = listent->next)
	{
		if (!strcmp (name, listent->name))
			return;
	}

	listent = (clist_t *)core_malloc (sizeof(clist_t));
	listent->name = core_strdup (name);
	String_Edit_To_Lower_Case (listent->name);

	//insert each entry in alphabetical order
    if (*headnode == NULL ||
	    strcasecmp(listent->name, (*headnode)->name) < 0) //insert at front
	{
        listent->next = *headnode;
        *headnode = listent;
    }
    else //insert later
	{
        prev = *headnode;
        cursor = (*headnode)->next;
		while (cursor && (strcasecmp(listent->name, cursor->name) > 0))
		{
            prev = cursor;
            cursor = cursor->next;
        }
        listent->next = prev->next;
        prev->next = listent;
    }
}


void List_Free (clist_t** headnode)
{
	clist_t *blah;

	while (*headnode)
	{
		blah = (*headnode)->next;
		core_free ((*headnode)->name);
		core_free (*headnode);

		*headnode = blah;
	}

	// headnode is NULL
}


int List_Count (clist_t *headnode)
{
	clist_t *cur;
	int count;

	for (cur = headnode, count = 0; cur; cur = cur->next, count ++);

	return count;
}

void List_Print (clist_t *headnode)
{
	clist_t *cur;
	int count;
	for (cur = headnode, count = 0; cur; cur = cur->next, count ++)
		Core_Printf ("%i: %s\n", count, cur->name);
}


clist_t *List_String_Split (const char *s, int ch_delimiter)
{
	clist_t *list = NULL;
	const char* linestart = s;
	const char* cursor = s;
	char *s2;
	size_t length =  0;

	for (linestart = s, cursor = s, length = 0; *cursor; cursor ++)
	{
		if (*cursor == ch_delimiter)
		{
			s2 = strndup (linestart, length);
			List_Add (&list, s2);
			free (s2);

			linestart = cursor + 1;
			length = 0;
		} else length++;
	}

	// Get the trailer
	s2 = strndup (linestart, length);
	List_Add (&list, s2);
	free (s2);
	
	return list;
}

